
##############################################################################
#
# namespace.tcl --
#
#    Introspection procedures to generate a Tcl script for initializing
#    new interpreters with namespaces and XOTcl object and classes.
#    This is extended version of procedures supplied by AOLserver.
#
#    See the file "license.txt" for information on usage and
#    redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
#    Rcsid: @(#)$Id: namespace.tcl,v 1.1 2001/06/17 09:27:36 zv Exp $
#
##############################################################################

#
#    If the namespace command does not exist, leave the simple global-only
#    procedure defined in C in place, which works fine for plain Tcl 7.6.
#

if {[info command namespace] == ""} {
	return
}


##############################################################################
#
# ::_ns_getnamespaces --
#
#   Obtains list of all namespaces present in the current interpreter.
#   Take care *not* to include XOTcl classes/objects (handled separately).
#
# Arguments:
#   nsplist        name of variable where to return result
#   topnsp         name of namespace to start with (optional, default = ::)
#
# Results:
#   None, but result left in the given variable.
#
##############################################################################

proc _ns_getnamespaces {nsplist {topnsp "::"}} {

    upvar $nsplist nsp
    lappend nsp $topnsp

    foreach n [namespace children $topnsp] {
        if {$n != "::XOTclClasses" && 
	    $n != "::xotcl" && 
	    $n != [info command $n]} {
            _ns_getnamespaces nsp $n
        }
    }
}

##############################################################################
#
# ::_ns_getcmds --
#
#   Obtains list of commands (procedures) in all namespaces.
#
# Arguments:
#   None.
#
# Results:
#   Script which can be used to regenerate all commands (procedures).
#
##############################################################################

proc _ns_getcmds {} {

    #
    # Get list of all namespaces to walk thru.
    #

    _ns_getnamespaces nsplist
    set cmds {}

    foreach n $nsplist {

        #
        # Cache found procedures there ...
        #

        foreach p [namespace eval $n {info procs *}] {
            if {$n == "::"} {
                set procs(::$p) 1
            } else {
                set procs(${n}::$p) 1
            }
        }

        #
        # so we can check and report only those which
        # are *not* imported from some other namespace.
        #

        foreach c [info commands ${n}::*] {
            if {[namespace origin $c] == $c} {
                if {![info exists procs($c)]} {
                    lappend cmds $c
                }
            }
        }
    }

    set cmds
}

##############################################################################
#
# ::_ns_getinit --
#
#   Return a script to initialize new interps. AOLserver will invoke this
#   script in the initial thread's interp at startup and after an ns_eval
#   to generate a script for initializing new connection interpreters.
#
# Arguments:
#   None.
#
# Results:
#   None. 
#
##############################################################################

proc _ns_getinit {} {

    set ncmds "" ; # namespace eval commands
    set icmds "" ; # namespace import commands
    set ecmds "" ; # namespace export commands

    array set procs [list] ; # tcl procedure names (fully qualified)

    ns_log notice "tcl: generating interp init script"
    _ns_getnamespaces nsplist

    #
    # Walk thru all namespaces and collect information about
    # procedures and variables found in each one of them.
    # For each item found, append commands to regenerate it.
    #

    foreach n $nsplist {

        append ncmds "namespace eval $n \{" \n

        #
        # Namespace own procedures
        #

        foreach p [namespace eval $n {info procs *}] {
            if {$n == "::"} {
                set fp ::$p
            } else {
                set fp ${n}::$p
            }
            set procs($fp) 1
            set o [namespace origin $fp]
            if {![string equal $fp $o]} {
                set nt [namespace tail $o]
                set no [namespace qual $o]
                append icmds "namespace import $o" \n
                append ecmds "namespace eval $no \"namespace export $nt\"" \n
            } else {
                set args "" 
                foreach a [info args $fp] {
                    if {[info default $fp $a d]} {
                        set a [list $a $d]
                    }
                    lappend args $a
                }
                append ncmds [list proc $p $args [info body $fp]] \n
            }
        }

        #
        # Import commands from other namespaces.
        # Original commands are already taken care by the C-code.
        #

        foreach c [info commands ${n}::*] {
            if {![info exists procs($c)]} {
                set o [namespace origin $c]
                if {![string equal $o $c]} {
                    append icmds "namespace import $origin" \n
                }
            }
        }

        #
        # Cover namespace own variables.
        #

        if {![string equal $n "::"]} {
            foreach v [info vars ${n}::*] {
                if {[info exists $v]} {
                    if {[array exists $v]} {
                        append ncmds [list variable $v] \n
                        append ncmds [list array set $v [array get $v]] \n
                    } else {
                        append ncmds [list variable $v [set $v]] \n
                    }
                }
            }
        }

        append ncmds "\}" \n
    }

    #
    # Conditionaly append XOTcl classes/objects definitions.
    #

    if {[info commands ::ScriptCreator] == "::ScriptCreator"} {
        ScriptCreator ScriptCreate
        append ncmds [ScriptCreate makeScriptForAll] \n
    } else {
        ns_log notice "XOTcl extension not loaded; will not copy objects."
    }

    append ncmds \n $ecmds \n $icmds

    if {0} {
        set f [open /tmp/_ns_getinit.tcl w]
        puts $f $ncmds
        close $f
        ns_log notice "tcl: WRITTEN TO /tmp/_ns_getinit.tcl"
    }


    return $ncmds
}

##############################################################################
#
# The part below is conditionaly executed only for the XOTcl extension.
# It copies class/object definitions from the startup to connection thread.
#
##############################################################################

if {[info comm ::Class] != "::Class" || [info comm ::Object] != "::Object"} {
    return
} else {
    set file [file join [file dirname [info script]] xotcl classcopy.xotcl]
    if {[file readable $file]} {
        source $file
    }
}

################################# End of file ################################
